--[[
	BActionButton
		A Bongos ActionButton
--]]

BActionButton = CreateFrame('CheckButton')
local Button_mt = {__index = BActionButton}

local SIZE = 36
local MAX_BUTTONS = 120 --the current maximum amount of action buttons
local buttons = {}
local showGrid


--[[ Button Events ]]--

local function OnUpdate() this:OnUpdate(arg1) end
local function PostClick() this:PostClick() end
local function OnDragStart() this:OnDragStart() end
local function OnReceiveDrag() this:OnReceiveDrag() end
local function OnEnter() this:OnEnter() end
local function OnLeave() this:OnLeave() end
local function OnShow() this:Update() end


--[[ Action Button Methods ]]--

BActionButton.HasNormalTexture = true

--Create an Action Button with the given ID and parent
function BActionButton.Create(id)
	if id > MAX_BUTTONS then return end

	local name = 'BActionButton' .. id
	local button = CreateFrame('CheckButton', name, nil, 'SecureActionButtonTemplate, ActionButtonTemplate')
	setmetatable(button, Button_mt)

	button:SetID(id)	
	button:RegisterForDrag("LeftButton", "RightButton")
	button:RegisterForClicks('anyUp')
	button:SetAttribute('type', 'action')
	button:SetAttribute('action', id)
	button:SetAttribute('useparent-unit', true)
	button:SetAttribute('checkselfcast', true)

	button:SetScript('OnUpdate', OnUpdate)
	button:SetScript('PostClick', PostClick)
	button:SetScript('OnDragStart', OnDragStart)
	button:SetScript('OnReceiveDrag', OnReceiveDrag)
	button:SetScript('OnEnter', OnEnter)
	button:SetScript('OnLeave', OnLeave)
	button:SetScript('OnShow', OnShow)

	button:Style()
	button:Hide()

	buttons[id] = button

	return button
end


--[[ attach and remove ]]--

function BActionButton.Set(id, parent)
	local button = buttons[id] or BActionButton.Create(id)

	parent:Attach(button)
	parent:SetAttribute('addchild', button)

	button:SetAttribute('bindings-shown', parent:GetKeys(id))
	button:ShowHotkey(BActionConfig.HotkeysShown())
	button:ShowMacro(BActionConfig.MacrosShown())

	return button
end

function BActionButton:Release()
	if InCombatLockdown() then SeeQ:Do(function() self:Release() end) return; end

	self:SetParent(nil)
	self:ClearAllPoints()
	self:SetAttribute('bindings-shown', nil)
	self:SetAttribute('showstates', nil)
	self:Hide()
end

function BActionButton:Style()
	local name = self:GetName()
	getglobal(name .. 'Icon'):SetTexCoord(0.06, 0.94, 0.06, 0.94)
	getglobal(name .. 'Border'):SetVertexColor(0, 1, 0, 0.6)
	getglobal(name .. 'NormalTexture'):SetVertexColor(1, 1, 1, 0.5)
end


--[[ OnX Functions ]]--

function BActionButton:OnUpdate(elapsed)
	local name = self:GetName()
	if not getglobal(name .. "Icon"):IsShown() then return end

	--update flashing
	if self.flashing == 1 then
		self.flashtime = self.flashtime - elapsed
		if self.flashtime <= 0 then
			local overtime = -self.flashtime
			if overtime >= ATTACK_BUTTON_FLASH_TIME then
				overtime = 0
			end
			self.flashtime = ATTACK_BUTTON_FLASH_TIME - overtime

			local flashTexture = getglobal(name .. "Flash")
			if flashTexture:IsShown() then
				flashTexture:Hide()
			else
				flashTexture:Show()
			end
		end
	end

	-- Handle range indicator
	if self.rangeTimer then
		if self.rangeTimer < 0 then
			local pagedID = self:GetPagedID()
			local hotkey = getglobal(name .. "HotKey")

			if IsActionInRange(pagedID) == 0 then
				hotkey:SetVertexColor(1, 0.1, 0.1)

				if BActionConfig.ColorOutOfRange() and IsUsableAction(pagedID) then
					local r,g,b = BActionConfig.GetRangeColor()
					getglobal(name .. "Icon"):SetVertexColor(r,g,b)
				end
			else
				hotkey:SetVertexColor(0.6, 0.6, 0.6)

				if IsUsableAction(pagedID) then
					getglobal(name .. "Icon"):SetVertexColor(1, 1, 1)
				end
			end
			self.rangeTimer = TOOLTIP_UPDATE_TIME
		else
			self.rangeTimer = self.rangeTimer - elapsed
		end
	end

	--Tooltip stuff, probably for the cooldown timer
	if self.updateTooltip then
		self.updateTooltip = self.updateTooltip - elapsed
		if self.updateTooltip <= 0 then
			if GameTooltip:IsOwned(self) then
				self:UpdateTooltip(self)
			else
				self.updateTooltip = nil
			end
		end
	end
end

function BActionButton:PostClick()
	self:UpdateState()
end

function BActionButton:OnDragStart()
	if not BActionConfig.ButtonsLocked() or (BActionConfig.IsQuickMoveKeyDown() or showGrid) then
		PickupAction(self:GetPagedID())
		self:UpdateState()
	end
end

function BActionButton:OnReceiveDrag()
	PlaceAction(self:GetPagedID())
	self:UpdateState()
end

function BActionButton:OnEnter()
	self:UpdateTooltip()
	KeyBound_Set(self)
end

function BActionButton:OnLeave()
	self.updateTooltip = nil
	GameTooltip:Hide()
end


--[[ Update Functions ]]--

--Updates the icon, count, cooldown, usability color, if the button is flashing, if the button is equipped,  and macro text.
function BActionButton:Update()
	if not self:GetParent() then return end

	local name = self:GetName()
	local action = self:GetPagedID()
	local icon = getglobal(name .. 'Icon')
	local cooldown = getglobal(name .. 'Cooldown')
	local texture = GetActionTexture(action)

	if texture then
		icon:SetTexture(texture); icon:Show()
		self.rangeTimer = -1
		
		if BActionButton.HasNormalTexture then
			self:SetNormalTexture('Interface\\Buttons\\UI-Quickslot2')
		end
	else
		icon:Hide()
		self.rangeTimer = nil
		cooldown:Hide()
		
		if BActionButton.HasNormalTexture then
			self:SetNormalTexture('Interface\\Buttons\\UI-Quickslot')
		end
		getglobal(name .. 'HotKey'):SetVertexColor(0.6, 0.6, 0.6)
	end

	if HasAction(action) then
		self:UpdateState()
		self:UpdateUsable()
		self:UpdateCooldown()
		self:UpdateFlash()
	else
		cooldown:Hide()
	end
	self:UpdateCount()

	-- Add a green border if button is an equipped item
	local border = getglobal(name .. 'Border')
	if IsEquippedAction(action) then
		border:SetVertexColor(0, 1, 0, 0.6)
		border:Show()
	else
		border:Hide()
	end

	if GameTooltip:IsOwned(self) then
		self:UpdateTooltip()
	else
		self.updateTooltip = nil
	end

	-- Update Macro Text
	getglobal(name .. 'Name'):SetText(GetActionText(action))
end

--Update the cooldown timer
function BActionButton:UpdateCooldown()
	local start, duration, enable = GetActionCooldown(self:GetPagedID())
	CooldownFrame_SetTimer(getglobal(self:GetName().."Cooldown"), start, duration, enable)
end

--Update item count
function BActionButton:UpdateCount()
	local text = getglobal(self:GetName() .. 'Count')
	local action = self:GetPagedID()

	if IsConsumableAction(action) then
		text:SetText(GetActionCount(action))
	else
		text:SetText('')
	end
end

--Update if a button is checked or not
function BActionButton:UpdateState()
	local action = self:GetPagedID()
	self:SetChecked(IsCurrentAction(action) or IsAutoRepeatAction(action))
end

--colors the action button if out of mana, out of range, etc
function BActionButton:UpdateUsable()
	local action = self:GetPagedID()
	local icon = getglobal(self:GetName().."Icon")

	local isUsable, notEnoughMana = IsUsableAction(action)
	if isUsable then
		if BActionConfig.ColorOutOfRange() and IsActionInRange(action) == 0 then
			local r,g,b = BActionConfig.GetRangeColor()
			icon:SetVertexColor(r,g,b)
		else
			icon:SetVertexColor(1, 1, 1)
		end
	elseif notEnoughMana then
		--Make the icon blue if out of mana
		icon:SetVertexColor(0.5, 0.5, 1)
	else
		--Skill unusable
		icon:SetVertexColor(0.3, 0.3, 0.3)
	end
end

function BActionButton:UpdateFlash()
	local action = self:GetPagedID()
	if (IsAttackAction(action) and IsCurrentAction(action)) or IsAutoRepeatAction(action) then
		self:StartFlash()
	else
		self:StopFlash()
	end
end

function BActionButton:StartFlash()
	self.flashing = 1
	self.flashtime = 0
	self:UpdateState()
end

function BActionButton:StopFlash()
	self.flashing = 0
	getglobal(self:GetName() .. 'Flash'):Hide()
	
	self:UpdateState()
end

function BActionButton:UpdateSlot()
	local changed = self:UpdateVisibility(BActionButton.ShowingEmpty())
	if changed then
		SecureStateHeader_Refresh(self:GetParent())
	else
		self:Update()
	end
end

--show if showing empty buttons, or if the slot has an action, hide otherwise
function BActionButton:UpdateVisibility(showEmpty)
	local oldstates = self:GetAttribute('showstates')
	local newstates

	if showEmpty or HasAction(self:GetPagedID()) then
		newstates = self:GetParent():GetShowStates()
	else
		newstates = '!*'
	end	
	
	if oldstates ~= newstates then
		self:SetAttribute('showstates', newstates)
		return true
	end
end

function BActionButton:UpdateTooltip()
	if BActionConfig.TooltipsShown() then
		if GetCVar("UberTooltips") == 1 then
			GameTooltip_SetDefaultAnchor(GameTooltip, self)
		else
			GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
		end

		local action = self:GetPagedID()
		if GameTooltip:SetAction(action) then
			self.updateTooltip = TOOLTIP_UPDATE_TIME
		else
			self.updateTooltip = nil
		end
	end
end


--[[ Hotkey Functions ]]--

function BActionButton:ShowHotkey(show)
	if show then
		getglobal(self:GetName() .. 'HotKey'):Show()
		self:UpdateHotkey()
	else
		getglobal(self:GetName() .. 'HotKey'):Hide()
	end
end

function BActionButton:UpdateHotkey()
	getglobal(self:GetName() .. 'HotKey'):SetText(self:GetHotkey() or '')
end

function BActionButton:GetHotkey()
	if self:GetParent() then
		local bindings = self:GetAttribute('bindings-shown')
		if bindings then
			return BActionUtil.ToShortKey(strsplit(';', bindings))
		end
	end
	return ''
end


--[[ Macro Functions ]]--

function BActionButton:ShowMacro(show)
	if show then
		getglobal(self:GetName() .. 'Name'):Show()
	else
		getglobal(self:GetName() .. 'Name'):Hide()
	end
end


--[[ Keybinding Functions ]]--

-- binds the given key to the given button
function BActionButton:SetKey(key)
	local parent = self:GetParent()

	local keys = parent:GetKeys(self:GetID())
	if keys then
		keys = keys .. key .. ';'
	else 
		keys = key .. ';'
	end
	parent:SetKeys(self:GetID(), keys, true)
end

local function RemoveKeyFromMap(key, map)
	if not(map and key) then return map end

	local newMap 
	local numKeys = select('#', strsplit(';', map))

	for i = 1, numKeys do
		local hotkey = select(i, strsplit(';', map))
		if not(hotkey == '' or hotkey == key) then
			if newMap then
				newMap = newMap .. hotkey .. ';'
			else
				newMap = hotkey .. ';'
			end
		end
	end
	return newMap
end

-- unbinds the given key from all other buttons
function BActionButton:FreeKey(key)
	for i = 1, BActionBar.GetNumber() do
		local bar = BActionBar.Get(i)
		local update
		for j = bar:GetStartID(), bar:GetEndID() do
			local keys = bar:GetKeys(j)
			if keys then
				update = true
				keys = RemoveKeyFromMap(key, keys)
				bar:SetKeys(j, keys)
			end
		end
		if update then
			bar:RefreshBindings()
		end
	end
end

-- removes all keys bound to the given button
function BActionButton:ClearBindings()
	self:GetParent():SetKeys(self:GetID(), nil, true)
end

-- returns a string listing all bindings of the given button
function BActionButton:GetBindings()
	local keys
	local bindings = self:GetAttribute('bindings-shown')
	if bindings then
		for i = 1, select('#', strsplit(';', bindings)) do
			local hotkey = select(i, strsplit(';', bindings))
			if hotkey ~= '' then
				if keys then
					keys = keys .. ',' .. GetBindingText(hotkey, 'KEY_')
				else
					keys = GetBindingText(hotkey, 'KEY_')
				end
			end
		end
	end
	return keys
end


--[[ Meta Functions ]]--

-- does the given action to every single button currently in use
function BActionButton.ForAll(action, ...)
	for _, button in pairs(buttons) do
		if button:GetParent() then
			action(button, ...)
		end
	end
end

-- does the given action to every single button being shown
function BActionButton.ForAllShown(action, ...)
	for _, button in pairs(buttons) do
		if button:GetParent() and button:IsShown() then
			action(button, ...)
		end
	end
end

--does the action to every single button being shown with an action
function BActionButton.ForAllWithAction(action, ...)
	for id, button in pairs(buttons) do
		if button:GetParent() and button:IsShown() and HasAction(id) then
			action(button, ...)
		end
	end
end

--does the action to every single button matching the given id
function BActionButton.ForID(id, action, ...)
	local button = buttons[id]
	if button and button:GetParent() then
		action(button, ...)
	end
end


--[[ Utility Functions ]]--

function BActionButton:GetPagedID()
	return self:GetID()
end

function BActionButton.Get(id)
	return buttons[id]
end

function BActionButton.GetMax()
	return MAX_BUTTONS
end

function BActionButton.GetSize()
	return SIZE
end

function BActionButton.ShowingEmpty()
	return showGrid or BActionConfig.ShowGrid()
end


--[[ Event Handler ]]--

local function OnEvent()
	if event == 'ACTIONBAR_SLOT_CHANGED' then
		local button = BActionButton.Get(arg1)
		if button and button:GetParent() then
			button:UpdateSlot()
		end
	elseif event == 'ACTIONBAR_SHOWGRID' then
		showGrid = true
		BActionBar.RefreshAll()
	elseif event == 'ACTIONBAR_HIDEGRID' then
		showGrid = nil
		BActionBar.RefreshAll()
	elseif event == 'PLAYER_ENTERING_WORLD' then
		BActionButton.ForAllShown(BActionButton.Update)
	elseif event == 'PLAYER_TARGET_CHANGED' or event == 'PLAYER_AURAS_CHANGED' then
		BActionButton.ForAllWithAction(BActionButton.UpdateUsable)
	elseif event == 'UNIT_INVENTORY_CHANGED' then
		if arg1 == 'player' then
			BActionButton.ForAllWithAction(BActionButton.Update)
		end
	elseif event == 'ACTIONBAR_UPDATE_USABLE' or event == 'UPDATE_INVENTORY_ALERTS' or event == 'ACTIONBAR_UPDATE_COOLDOWN' then
		BActionButton.ForAllWithAction(BActionButton.UpdateCooldown)
		BActionButton.ForAllWithAction(BActionButton.UpdateUsable)
	elseif event == 'ACTIONBAR_UPDATE_STATE' or event == 'CRAFT_SHOW' or event == 'CRAFT_CLOSE' or event == 'TRADE_SKILL_SHOW' or event == 'TRADE_SKILL_CLOSE' then
		BActionButton.ForAllWithAction(BActionButton.UpdateState)
	elseif event == 'PLAYER_ENTER_COMBAT' or event == 'PLAYER_LEAVE_COMBAT' or event == 'START_AUTOREPEAT_SPELL' or event == 'STOP_AUTOREPEAT_SPELL' then
		BActionButton.ForAllWithAction(BActionButton.UpdateFlash)
	elseif event == 'UPDATE_BINDINGS' then
		BActionBar.RefreshAllBindings()
	end
end

local function RegisterEvents()
	local frame = CreateFrame("Frame")

	frame:SetScript('OnEvent', OnEvent)
	frame:RegisterEvent("ACTIONBAR_SHOWGRID")
	frame:RegisterEvent("ACTIONBAR_HIDEGRID")	
	frame:RegisterEvent("ACTIONBAR_SLOT_CHANGED")
	frame:RegisterEvent('ACTIONBAR_UPDATE_STATE')
	frame:RegisterEvent('ACTIONBAR_UPDATE_USABLE')
	frame:RegisterEvent('ACTIONBAR_UPDATE_COOLDOWN')
	frame:RegisterEvent('UPDATE_INVENTORY_ALERTS')
	frame:RegisterEvent('PLAYER_AURAS_CHANGED')
	frame:RegisterEvent('UNIT_INVENTORY_CHANGED')
	frame:RegisterEvent('CRAFT_SHOW')
	frame:RegisterEvent('CRAFT_CLOSE')
	frame:RegisterEvent('TRADE_SKILL_SHOW')
	frame:RegisterEvent('TRADE_SKILL_CLOSE')
	frame:RegisterEvent('PLAYER_ENTER_COMBAT')
	frame:RegisterEvent('PLAYER_LEAVE_COMBAT')
	frame:RegisterEvent('START_AUTOREPEAT_SPELL')
	frame:RegisterEvent('STOP_AUTOREPEAT_SPELL')
	frame:RegisterEvent('UPDATE_BINDINGS')

	BVent:AddAction('KEYBOUND_ENABLED', function() BActionBar.RefreshAll(true) end)
	BVent:AddAction('KEYBOUND_DISABLED', function() BActionBar.RefreshAll(nil) end)
end


--[[ Startup ]]--

local eventsLoaded
Bongos.AddStartup(function()
	if not eventsLoaded then 
		RegisterEvents()
		eventsLoaded = true 
	end
end)